Package: cpfa
Type: Package
Title: Classification with Parallel Factor Analysis
Version: 1.3.1
Date: 2026-07-13
Authors@R: person(given = c("Matthew", "A."), family = "Asisgress", 
                  email = "mattgress@protonmail.ch", role = c("aut", "cre"), 
                  comment = c(ORCID = "0000-0001-7294-3348"))
Maintainer: Matthew A. Asisgress <mattgress@protonmail.ch>
Depends: R (>= 4.3.0), multiway
Imports: glmnet, e1071, randomForest, nnet, rda, xgboost, foreach,
        doParallel, doRNG, clue
Suggests: knitr, rmarkdown
Description: Classification using Richard A. Harshman's Parallel Factor 
    Analysis-1 (Parafac) model or Parallel Factor Analysis-2 (Parafac2) model fit to 
    a three-way or four-way data array. See Harshman and Lundy (1994): 
    <doi:10.1016/0167-9473(94)90132-5>. Classification using principal component 
    analysis (PCA) fit to a two-way data matrix is also supported. Uses component 
    weights from one mode of a Parafac, Parafac2, or PCA model as features to tune 
    parameters for one or more classification methods via a k-fold cross-validation 
    procedure. Allows for constraints on different tensor modes. Allows for 
    inclusion of additional features alongside features generated by the component 
    model. Supports penalized logistic regression, support vector machine, random 
    forest, feed-forward neural network, regularized discriminant analysis, and 
    gradient boosting machine. Supports binary and multiclass classification. 
    Predicts class labels or class probabilities and calculates multiple 
    classification performance measures. Uses the 'clue' package to align Parafac 
    or Parafac2 models across data splits in the cross-validation procedure. 
    Calculates classification importance of individual features using permutation 
    feature importance. Implements parallel computing via the 'foreach', 
    'doParallel', and 'doRNG' packages.
BugReports: https://github.com/matthewasisgress/cpfa/issues
License: GPL (>= 2)
URL: https://github.com/matthewasisgress/cpfa
VignetteBuilder: knitr, rmarkdown
Encoding: UTF-8
NeedsCompilation: no
Packaged: 2026-07-13 00:03:09 UTC; nr2
Author: Matthew A. Asisgress [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-7294-3348>)
Repository: CRAN
Date/Publication: 2026-07-13 05:10:08 UTC
Built: R 4.6.1; ; 2026-07-13 09:32:08 UTC; unix
