exportPattern("^[[:alpha:]]+")

import(multiway)
import(glmnet)
import(e1071)
import(randomForest)
import(nnet)
import(rda)
import(xgboost)
import(doParallel)
import(doRNG)
import(clue)

S3method(print, tunecpfa)
S3method(predict, tunecpfa)

importFrom(stats, "na.omit", "predict", "model.matrix", "var", "formula", "cor", "varimax", "sd")
importFrom(stats, "rbinom", "rexp", "rgamma", "rbeta", "rnorm", "rpois")
importFrom(stats, "runif", "rgeom", "rnbinom", "rhyper", "rlnorm", "rcauchy")
importFrom(stats, "dnorm", "qlogis", "qnorm", "quantile", "median")
importFrom(utils, "combn")
importFrom("graphics", "axis", "image", "layout", "mtext", "par", "boxplot", "box", "text")
importFrom(parallel, "makeCluster", "detectCores", "clusterEvalQ", "stopCluster", "clusterSetRNGStream")
importFrom("grDevices", "colorRampPalette", "devAskNewPage")
importFrom(foreach, "%do%", "%dopar%", "foreach")
