\name{cpfa-internals}

\title{Internal Functions for "cpfa"}

% List of Internal Functions
\alias{kcv.plr}
\alias{kcv.svm}
\alias{kcv.rf}
\alias{kcv.nn}
\alias{kcv.rda}
\alias{kcv.gbm}
\alias{distdraw}
\alias{kcvcheck}
\alias{logicheck}
\alias{numcheck}

\description{
  Internal functions for the "cpfa" package.
}
\details{
  These functions are not intended to be called by the user.
}
\keyword{internal}